--CPML is the magic that makes the transforms work
local cpml = require "cpml"
local modelLoader = require "modelLoader"

local gun = {}
local egg = {}
local cube = {}
local shader, matrix, makeProper, drawModel

--Function to correct cpml incorrect matrix format.
function makeProper(mat)
	local m = cpml.mat4.new()
	return cpml.mat4.transpose(m, mat)
end

love.graphics.setDefaultFilter("nearest", "nearest")

--initial position and angle of Camera
camera = {
	pos 			= cpml.vec3(0, 0, -15),
	angle			= cpml.vec3(0, 0, 0),
	perspective 	= makeProper(cpml.mat4.from_perspective(60, love.graphics.getWidth()/love.graphics.getHeight(), 0.1, 1000)),
	transform 		= cpml.mat4(),
}

--initial transforms for the model
local gunTransform = cpml.mat4.identity()

local eggTransform = cpml.mat4.identity()
eggTransform:translate(eggTransform, cpml.vec3(12,0,0))

local cubeTransform = cpml.mat4.identity()
cubeTransform:translate(cubeTransform, cpml.vec3(-12,0,0))

--Custom vertex format add z, w to vertex position and the normals for vertices
local format = { {"VertexPosition","float",4}, {"VertexTexCoord","float",2}, {"VertexNormal","float",3} }

function love.load()
	love.graphics.setBackgroundColor(.5,.5,.5)

	local gunMesh = love.graphics.newMesh(format, modelLoader("/models/egg.obj"), "triangles")
	gunMesh:setTexture(love.graphics.newImage("/images/egg.png"))

	gun = {
		transform = gunTransform,
		mesh = gunMesh
	}

	local eggMesh = love.graphics.newMesh(format, modelLoader("/models/egg.obj"), "triangles")
	eggMesh:setTexture(love.graphics.newImage("/images/egg.png"))

	egg = {
		transform = eggTransform,
		mesh = eggMesh
	}

	local cubeMesh = love.graphics.newMesh(format, modelLoader("/models/cube.obj"), "triangles")
	cubeMesh:setTexture(love.graphics.newImage("/images/boxtexture.png"))

	cube = {
		transform = cubeTransform,
		mesh = cubeMesh
	}

	shader = love.graphics.newShader("basicRender.glsl")

	--Sets DepthMode. Prevents items being drawn out of order
	love.graphics.setDepthMode("lequal", true)

end

function love.draw()
	startDraw()
		drawModel(gun)
		drawModel(egg)
		drawModel(cube)
	endDraw()

	love.graphics.print(string.format("Angle: %.2f, %.2f", camera.angle.x, camera.angle.y), 20, 20)
	love.graphics.print(string.format("Position: %.2f, %.2f, %.2f", camera.pos.x, camera.pos.y, camera.pos.z), 20, 20*2)
	love.graphics.print("Backspace to reset position and angle", 20, 20*3)
	love.graphics.print("Click TAB to lock mouse", 20, 20*4)
	love.graphics.print("WASD to move camera position", 20, 20*5)
	love.graphics.print("FPS " .. love.timer.getFPS(), 20, 20*6)
end

function love.update(dt)
	local speed = 5 * dt
	local pos = camera.pos
	
	local mul = love.keyboard.isDown("w", "up") and 1 or (love.keyboard.isDown("s", "down") and -1 or 0)
	pos.x = pos.x + math.sin(-camera.angle.x) * mul * speed
	pos.z = pos.z + math.cos(-camera.angle.x) * mul * speed
	
	local mul = love.keyboard.isDown("d", "right") and -1 or (love.keyboard.isDown("a", "left") and 1 or 0)
	pos.x = pos.x + math.cos(camera.angle.x) * mul * speed
	pos.z = pos.z + math.sin(camera.angle.x) * mul * speed

	local mul = love.keyboard.isDown("e") and 1 or (love.keyboard.isDown("q") and -1 or 0)
	pos.y = pos.y + mul * speed
end

--Function to merge camera transform and projection matrix(I think anyways)
function setCamera()
	camera.transform = cpml.mat4() -- Wipe transform
	local t, a, p = camera.transform, camera.angle, camera.pos

	t:rotate(t, a.y, cpml.vec3.unit_x)
	t:rotate(t, a.x, cpml.vec3.unit_y)
	t:rotate(t, a.z, cpml.vec3.unit_z)
	
	t:translate(t, p) -- Rotation and then translating makes the rotation relative to camera, rather than origin
	
	return camera.perspective * makeProper(t)
end

function love.keypressed(key)
	if key == "escape" then
		love.event.quit()
	elseif key == "backspace" then
		camera.angle = cpml.vec3(0, 0, 0)
		camera.pos = cpml.vec3(0, 0, -15)
	elseif key == "tab" then
		love.mouse.setRelativeMode(not love.mouse.getRelativeMode())
	end
end

function love.mousemoved(x, y, dx, dy)
	if love.mouse.getRelativeMode() then
		camera.angle.x = camera.angle.x + math.rad(dx * 0.5)
		camera.angle.y = camera.angle.y + math.rad(dy * 0.5)
	end
end

function startDraw()
	love.graphics.push("all")
	--centers the world on the screen
	local w, h = love.graphics.getDimensions()
	love.graphics.translate(w/2, h/2)

	--Sends the view and model transforms to shader
	local view = setCamera()
	shader:send("view", view)
	
	love.graphics.setShader(shader)
end

function drawModel(model)
	shader:send("model_matrix", makeProper(model.transform))
	love.graphics.draw(model.mesh)
end

function endDraw()
	love.graphics.setShader()
	love.graphics.pop()
end
